from os import path

from canaryd_packages import six
from canaryd_packages.six.moves import shlex_quote

from canaryd.plugin import Plugin
from canaryd.script import get_scripts, get_scripts_directory
from canaryd.subprocess import CalledProcessError, get_command_output


class Scripts(Plugin):
    '''
    The scripts plugin executes user enabled scripts.

    Scripts have the following spec (think Sensu/Nagios):

    + Output a message and/or JSON data to stdout or stderr
    + Exit with code:
        * 0 - indicates all OK
        * 1 - indicates a warning
        * 2 - indicates a critical

    Scripts can be added to the canaryd config directory in ``scripts/available``
    and then enabled using ``canaryctl``, eg:

    .. code:: sh

        # Enable/disable scripts in $canaryd-config/scripts/available
        canaryctl script enable FILENAME
        canaryctl script disable FILENAME

        # Create a new script
        canaryctl script create FILENAME

        # List all scripts (available and enabled)
        canaryctl scripts
    '''

    spec = ('script', {
        'output': six.text_type,
        'exit_code': int,
    })

    # Don't generate events from state updates (see generate_issues_from_key_change below)
    generate_update_events = False

    def prepare(self, settings):
        pass

    def get_state(self, settings):
        scripts = [
            (script, path.join(get_scripts_directory(), 'enabled', script))
            for script, enabled in get_scripts(settings)
            if enabled
        ]

        results = {}

        for script, script_path in scripts:
            try:
                output = get_command_output(
                    shlex_quote(script_path),
                )

                results[script] = {
                    'output': output.strip(),
                    'exit_code': 0,
                }

            except (CalledProcessError, OSError) as e:
                results[script] = {
                    'output': e.output.strip(),
                    'exit_code': e.returncode,
                }

        return results

    @staticmethod
    def generate_issues_from_change(change, settings):
        data_changes = change.data

        # If the script has been removed, resolve any leftover issues and exit.
        # (the delete event is still created).
        if change.type == 'deleted':
            yield 'resolved', None, None
            return

        if 'exit_code' in data_changes:
            _, to_code = data_changes['exit_code']

            message = None
            data = data_changes

            if 'output' in data_changes:
                data = dict(
                    (k, v)
                    for k, v in six.iteritems(data_changes)
                    if k != 'output'
                )
                message = data_changes['output'][1]

            if to_code >= 2:
                yield 'critical', message, data

            elif to_code == 1:
                yield 'warning', message, data

            elif to_code == 0:
                # Include all the changes - including output
                yield 'resolved', message, data_changes



from __future__ import division

from collections import defaultdict, deque
from itertools import islice
from time import sleep, time

from canaryd_packages import six

from canaryd.log import logger
from canaryd.plugin import Plugin
from canaryd.subprocess import CalledProcessError

from .monitor_util import get_cpu_stats, get_disk_stats, get_memory_stats


def sleep_until_interval(start, interval):
    time_taken = time() - start

    if time_taken < interval:
        sleep(interval - time_taken)


class Monitor(Plugin):
    '''
    This plugin monitors the servers basics (CPU/memory/disk/etc) and provides
    warning/critical alerts as configured on ``app.servicecanary.com``.

    Tracks:

    + CPU/IO wait usage %
    + Memory/swap usage %
    + Per disk usage %

    The actual tracking happens in a separate thread.
    '''

    spec = ('key', {
        'type': six.text_type,

        # The value, max and the % used
        'percentage': float,

        # New (>=0.3) replacements for above
        'current_value': (int, 'long'),
        'current_max': (int, 'long'),

        # Rolling time max/mins
        '1_min_max_percentage': float,
        '1_min_min_percentage': float,
        '5_min_max_percentage': float,
        '5_min_min_percentage': float,
        '15_min_max_percentage': float,
        '15_min_min_percentage': float,

        # Rolling percentage averages
        '1_min_percentage': float,
        '5_min_percentage': float,
        '15_min_percentage': float,

        # Legacy support (<0.3)
        'value': int,
        'max': int,
    })

    # Disable update diff-ing - meaning generate_events always sees the full
    # state (the full state is always sent to the server). This is needed because
    # we need to check for both critical and warning, which might have different
    # average timescales - so we always need value and 1/5/15 min averages.
    diff_updates = False

    # Don't generate events from state updates (see generate_issues_from_key_change below)
    generate_update_events = False

    # Don't log warnings when keys are missing (eg 15_min_percentage)
    warn_for_missing_keys = False

    collect_interval = None

    def setup_history(self):
        # Number of items needed if one per collect interval to total 15 mins
        intervals = int(round((15 * 60) / self.collect_interval))
        logger.info('Setup interval history with {0} slots'.format(intervals))

        self.history = defaultdict(lambda: deque((), intervals))

    def prepare(self, settings):
        if self.collect_interval != settings.collect_interval_s:
            self.collect_interval = settings.collect_interval_s
            self.setup_history()

    def get_stats(self):
        seen_keys = set()

        for type_, collector in (
            ('cpu', get_cpu_stats),
            ('memory', get_memory_stats),
            ('disk', get_disk_stats),
        ):
            try:
                stats = collector()

            except (CalledProcessError, OSError) as e:
                logger.warning('Error collecting {0} stats: {1}'.format(type_, e))
                continue

            # Push all the stats -> history
            for key, data in six.iteritems(stats):
                data['type'] = type_
                data['percentage'] = round(data['percentage'], 2)
                self.history[key].appendleft(data)
                seen_keys.add(key)

        # Remove anything not found in the latest stats
        history_keys = set(six.iterkeys(self.history))
        for key_to_remove in history_keys - seen_keys:
            self.history.pop(key_to_remove)

    def get_state(self, settings):
        # If the collect interval has changed, reset history otherwise we'll
        # generate bad averages.
        if self.collect_interval != settings.collect_interval_s:
            self.collect_interval = settings.collect_interval_s
            self.setup_history()

        # Get all the latest stats
        self.get_stats()

        intervals = {
            '1_min': max(60 / self.collect_interval, 1),
            '5_min': max((5 * 60) / self.collect_interval, 1),
            '15_min': max((15 * 60) / self.collect_interval, 1),
        }

        data = {}

        for key, history in six.iteritems(self.history):
            if not history:
                continue

            data[key] = history[0]

            for interval_key, interval_length in six.iteritems(intervals):
                if len(history) < interval_length:
                    continue

                items = [
                    item['percentage']
                    for item in islice(history, int(interval_length))
                ]

                percentage_key = '{0}_percentage'.format(interval_key)
                data[key][percentage_key] = round(sum(items) / interval_length, 2)

                min_key = '{0}_min_percentage'.format(interval_key)
                data[key][min_key] = min(items)

                max_key = '{0}_max_percentage'.format(interval_key)
                data[key][max_key] = max(items)

        return data

    @staticmethod
    def generate_issues_from_change(change, settings):
        key = change.key
        event_type = change.type
        data_changes = change.data

        settings_key = key

        # Swap is treated like memory (>X% warning/critical)
        if key == 'swap':
            settings_key = 'memory'

        # Anything not cpu/memory/iowait is assumed a disk
        elif key not in ('cpu', 'memory', 'iowait'):
            settings_key = 'disk'

        def make_event(alert_type, type_, limit, time, changes):
            message_key = key

            if settings_key == 'cpu':
                message_key = 'CPU'

            elif settings_key != 'disk':
                message_key = message_key.title()

            # Time=0, instant alert
            message = '{0} is over {1}%'.format(message_key, limit)

            time_to_text = {
                60: '1 minute',
                300: '5 minutes',
                900: '15 minutes',
            }

            if time and time > 0:
                time_text = time_to_text[time]

                if type_ == 'always':
                    message = '{0} over {1}% for {2}'.format(
                        message_key, limit, time_text,
                    )

                elif type_ == 'once':
                    message = '{0} over {1}% in the last {2}'.format(
                        message_key, limit, time_text,
                    )

                else:
                    message = '{0} average over {1}% for {2}'.format(
                        message_key, limit, time_text,
                    )

            if limit is None:
                message = '{0} is back to normal'.format(message_key)

            return alert_type, message, changes

        # If the item has been removed, resolve any lefover issues and exit
        if event_type == 'deleted':
            yield make_event('resolved', None, None, None, None)
            return

        time_setting_to_key = {
            0: 'percentage',
            60: '1_min',
            300: '5_min',
            900: '15_min',
        }

        # Status of over-limit values (is_warning, was_warning)
        resolved_changes = {}
        wanted_data_keys = set()

        for alert_type in ('critical', 'warning'):
            enabled = getattr(
                settings,
                '{0}_{1}'.format(settings_key, alert_type),
            )

            if not enabled:
                continue

            # The length of time to be over
            time = getattr(
                settings,
                '{0}_{1}_time_s'.format(settings_key, alert_type),
            )

            # Invalid time? Ignore it!
            if time not in time_setting_to_key:
                continue

            data_key = time_setting_to_key[time]

            # Always (default), average or once
            type_ = getattr(
                settings,
                '{0}_{1}_type'.format(settings_key, alert_type),
            )

            # If we're instant, use the latest/current percentage as-is
            if time > 0:
                type_to_key_formatter = {
                    'always': '{0}_min_percentage',
                    'once': '{0}_max_percentage',
                    'average': '{0}_percentage',
                }

                new_data_key = type_to_key_formatter[type_].format(data_key)

                # COMPAT w/<0.3 (no X_[min|max]_percentage)
                # Default to the rolling average as that's the only thing supported
                if new_data_key not in data_changes:
                    type_ = 'average'
                    data_key = '{0}_percentage'.format(data_key)
                else:
                    data_key = new_data_key

            # The limit to go over
            limit = getattr(
                settings,
                # Eg cpu_warning_limit
                '{0}_{1}_limit'.format(settings_key, alert_type),
            )

            wanted_data_keys.add(data_key)

            if data_key in data_changes:
                old_value, value = data_changes[data_key]

                if value > limit:
                    yield make_event(alert_type, type_, limit, time, {
                        data_key: data_changes[data_key],
                    })
                    # This works because critical is run before warning
                    return

                resolved_changes[data_key] = data_changes[data_key]

        if all(k in resolved_changes for k in wanted_data_keys):
            yield make_event('resolved', None, None, None, resolved_changes)



# canaryd
# File: canaryd/plugin.py
# Desc: functions for handling the various built in state gathering plugins

from __future__ import division

import re
import traceback

from distutils.spawn import find_executable
from glob import glob
from os import path

from canaryd_packages import six
from canaryd_packages.importlib import import_module

from canaryd.log import logger, print_exception
from canaryd.subprocess import get_command_output

PLUGINS = []
NAME_TO_PLUGIN = {}


def underscore(name):
    '''
    Transform CamelCase -> snake_case.
    '''

    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


class NoPluginError(Exception):
    pass


class PluginMeta(type):
    def __init__(cls, name, bases, attrs):
        if name == 'Plugin':
            return

        name = underscore(name)

        plugin_instance = cls()
        plugin_instance.name = name

        PLUGINS.append(plugin_instance)
        NAME_TO_PLUGIN[name] = plugin_instance


@six.add_metaclass(PluginMeta)
class Plugin(object):
    diff_updates = True

    generate_update_events = True
    generate_added_events = True
    generate_deleted_events = True

    # Log warnings for missing keys when validating state items?
    warn_for_missing_keys = True

    class PrepareError(Exception):
        pass

    def __repr__(self):
        return 'Plugin: {0}'.format(self.name)

    @staticmethod
    def get_action_for_change(change):
        pass

    @staticmethod
    def get_description_for_change(change):
        pass

    @staticmethod
    def get_change_key(data_changes):
        return ''

    @staticmethod
    def should_apply_change(change):
        return True

    @staticmethod
    def generate_issues_from_change(change, settings):
        pass

    def get_timeout(self, settings):
        # Set an alarm: a plugin can only run for MAX half the interval time
        return int(round(settings.collect_interval_s / 2))

    def get_state(self, settings):
        data = get_command_output(
            self.command,
            timeout=self.get_timeout(settings),
        )

        return self.parse(data)

    def prepare(self, settings):
        command_bits = self.command.split()
        command = command_bits[0]

        if not find_executable(command):
            raise OSError('Could not find executable: {0}'.format(command))

    def log(self, message):
        message = '[{0}]: {1}'.format(self.name, message)
        logger.debug(message)

    def check_spec_key(self, key, value):
        spec = self.spec[1]

        if key not in spec:
            raise ValueError(
                'Key not found in plugin {0} spec: {1} (value = {2})'.format(
                    self.name,
                    key,
                    value,
                ),
            )

        wanted_type = spec[key]
        valid = False

        def validate_type(value, wanted_type):
            return (
                # None/null allowed for all keys
                value is None
                # Or the wanted type
                or isinstance(value, wanted_type)
                # If we wan't unicode but got bytes, OK to serialise
                or (
                    wanted_type is six.text_type
                    and isinstance(value, six.binary_type)
                )
            )

        # Any type allowed?
        if wanted_type is None:
            valid = True

        # If the spec is a list - type is/should be the only list item, and our
        # value should be a list of items that match the type.
        elif isinstance(wanted_type, list):
            wanted_type = wanted_type[0]

            if isinstance(value, (list, set, tuple)) and all(
                validate_type(v, wanted_type)
                for v in value
            ):
                valid = True

        elif validate_type(value, wanted_type):
            valid = True

        if not valid:
            raise TypeError((
                'Invalid type for key `{0}` in plugin {1} '
                '(want type = {2}, value type = {3})'
            ).format(key, self.name, wanted_type, type(value)))

    def validate_state(self, state):
        '''
        Validate a given state dict is valid by checking each key/value within
        an item match up with the spec.
        '''

        spec = self.spec[1]
        key_to_missing_count = {}

        for key, item in six.iteritems(state):
            # Check all the keys the item has
            for k, v in six.iteritems(item):
                self.check_spec_key(k, v)

            # Check, but only warn, for keys the item *doesn't* have - this is
            # only because it's useful, plugins are free to submit incomplete
            # items.
            for spec_key in spec.keys():
                if spec_key not in item:
                    key_to_missing_count.setdefault(spec_key, 0)
                    key_to_missing_count[spec_key] += 1

        # Not all plugins expect all keys
        if not self.warn_for_missing_keys:
            return

        for key, missing_count in six.iteritems(key_to_missing_count):
            logger.warning('{0} items for plugin {1} are missing key {2}'.format(
                missing_count,
                self.name,
                key,
            ))


def get_plugins():
    '''
    Get the list of installed plugins.
    '''

    module_filenames = glob(path.join(path.dirname(__file__), 'plugins', '*.py'))

    module_names = [
        path.basename(name)[:-3]
        for name in module_filenames
        if not name.endswith('__init__.py')
    ]

    # Import all the modules to populate PLUGINS
    for name in module_names:
        import_module('canaryd.plugins.{0}'.format(name))

    return PLUGINS


def get_plugin_names():
    return six.iterkeys(NAME_TO_PLUGIN)


def get_plugin_by_name(plugin_name):
    if plugin_name not in NAME_TO_PLUGIN:
        raise NoPluginError('Missing plugin: {0}'.format(plugin_name))

    return NAME_TO_PLUGIN[plugin_name]


def prepare_plugin(plugin, settings):
    logger.debug('Preparing plugin: {0}'.format(plugin))

    try:
        plugin.prepare(settings)

    except Plugin.PrepareError as e:
        logger.info('Plugin prepare failed: {0}: {1}'.format(
            plugin.name, e.message,
        ))
        print_exception(debug_only=True)
        return False, e

    except OSError as e:
        logger.info('Plugin command missing/failed: {0}'.format(
            plugin.name,
        ))
        print_exception(debug_only=True)
        return False, e

    except Exception as e:
        logger.warning('Error preparing plugin: {0}: {1}'.format(
            plugin.name,
            e,
        ))
        print_exception()
        return False, e

    return True


def get_and_prepare_working_plugins(settings):
    all_plugins = get_plugins()
    working_plugins = []

    for plugin in all_plugins:
        status = prepare_plugin(plugin, settings)

        if status is True:
            working_plugins.append(plugin)

    logger.info('Loaded plugins: {0}'.format(', '.join([
        plugin.__class__.__name__
        for plugin in working_plugins
    ])))

    return working_plugins


def get_plugin_state(plugin, settings):
    '''
    Gets state output from a single plugin.
    '''

    logger.debug('Running plugin: {0}'.format(plugin))

    status = False
    state = None

    try:
        state = plugin.get_state(settings)
        status = True

    except Exception as e:
        logger.warning('Error running plugin: {0}: {1}'.format(plugin, e))
        print_exception()
        # Attach the tracback to the exception before returning as state
        e._traceback = traceback.format_exc()
        state = e

    # Validate the state
    if status:
        plugin.validate_state(state)

    return status, state


def get_plugin_states(plugins, settings):
    '''
    Gets state output from multiple plugins.
    '''

    return [
        (plugin, get_plugin_state(plugin, settings))
        for plugin in plugins
    ]



#!/usr/bin/env python

import sys

from os import path


# If /proc/mdstat doesn't exist this script has nothing to check!
if not path.exists('/proc/mdstat'):
    print('No /proc/mdstat found!')
    sys.exit(1)

# Load /proc/mdstat
with open('/proc/mdstat', 'r') as f:
    mdadm_data = f.read()


warnings = []
criticals = []

device = None


for line in mdadm_data.splitlines():
    if not line:
        continue

    if not line.startswith(' '):
        bits = line.split()
        device = bits[0]

        # Handle failed drives in the definition line
        for bit in bits:
            if bit.endswith('(F)'):
                criticals.append((
                    'RAID device {0} has a failed drive: {1}'
                ).format(device, bit))

    # Handle the second line of a device
    if device and 'blocks' in line:
        device_counts = line.split()[-2][1:-1]
        wanted_devices, devices = device_counts.split('/')

        if wanted_devices != devices:
            criticals.append((
                'RAID device {0} is missing devices: {1}/{2}'
            ).format(device, devices, wanted_devices))


exit_code = 0

if warnings:
    exit_code = 1
    print('\n'.join(warnings))

if criticals:
    exit_code = 2
    print('\n'.join(criticals))

if exit_code == 0:
    print('All RAID devices healthy')


sys.exit(exit_code)



import os
import shlex
import sys

from canaryd_packages import six
# Not ideal but using the vendored in (to requests) chardet package
from canaryd_packages.requests.packages import chardet

from canaryd.log import logger


if os.name == 'posix' and sys.version_info[0] < 3:
    from canaryd_packages.subprocess32 import *  # noqa
else:
    from subprocess import *  # noqa


def get_command_output(command, *args, **kwargs):
    logger.debug('Executing command: {0}'.format(command))

    if (
        not kwargs.get('shell', False)
        and not isinstance(command, (list, tuple))
    ):
        command = shlex.split(command)

    output = check_output(  # noqa
        command,
        close_fds=True,
        stderr=STDOUT,  # noqa
        *args, **kwargs
    )

    if isinstance(output, six.binary_type):
        encoding = chardet.detect(output)['encoding']
        if encoding:
            output = output.decode(encoding)
        else:
            output = output.decode()

    return output



import socket

from json import dumps as json_dumps, JSONEncoder
from time import sleep

from canaryd_packages import requests

from canaryd.log import logger
from canaryd.settings import get_settings
from canaryd.version import __version__

SESSION = None
REQUEST_TIMEOUT = 30


class CanaryJSONEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)

        if hasattr(obj, 'serialise'):
            return obj.serialise()

        if hasattr(obj, 'isoformat'):
            return obj.isoformat()

        return JSONEncoder.default(self, obj)


def backoff(function, *args, **kwargs):
    data = None
    interval = 0

    error_message = kwargs.pop('error_message', 'API error')
    max_wait = kwargs.pop('max_wait', 300)

    while data is None:
        try:
            return function(*args, **kwargs)

        except ApiError as e:
            if interval + 10 <= max_wait:
                interval += 10

            e.log()

        logger.critical('{0}, retrying in {1}s'.format(error_message, interval))
        sleep(interval)


class ApiError(Exception):
    '''
    Generic exception for problems with the API.
    '''

    def __init__(self, status_code, name, message=None, content=None):
        self.status_code = status_code
        self.name = name
        self.message = message
        self.content = content

    def log(self):
        logger.critical(
            '{0}: {1}{2}'.format(
                self.status_code,
                self.name,
                '({0})'.format(self.message) if self.message else '',
            ),
        )

        if self.content:
            logger.debug('Response data: {0}'.format(self.content))


def get_session():
    global SESSION

    if not SESSION:
        SESSION = requests.Session()

    return SESSION


def make_sync_or_changes_dict(states):
    plugin_to_state = {}

    for plugin, (status, state) in states:
        # Plugin ran OK - just attach any state
        if status is True:
            plugin_to_state[plugin.name] = state

        # Plugin ran OK but needs to SYNC with the server, sending a full state
        # or the plugin failed to run, send exception data
        elif status in ('SYNC', 'ERROR'):
            plugin_to_state[plugin.name] = (status, state)

        else:
            raise TypeError('Unknown status for {0} plugin: {1}/{2}'.format(
                plugin.name, status, state,
            ))

    return plugin_to_state


def make_api_request(
    method, endpoint,
    api_key=None, settings=None, json=None, **kwargs
):
    settings = settings or get_settings()

    api_key = api_key or settings.api_key

    url = '{0}/v{1}/{2}'.format(
        settings.api_base,
        settings.api_version,
        endpoint,
    )

    logger.debug('Making API request: {0}'.format(url))

    if 'timeout' not in kwargs:
        kwargs['timeout'] = REQUEST_TIMEOUT

    if json:
        json_data = json_dumps(json, cls=CanaryJSONEncoder)
        kwargs['data'] = json_data
        kwargs['headers'] = {
            'Content-Type': 'application/json',
        }

        logger.debug('Request data: {0}'.format(json_data))

    try:
        response = method(
            url,
            auth=('api', api_key),
            **kwargs
        )

    # Connection errors and timeouts
    except requests.ConnectionError as e:
        raise ApiError(0, 'Could not connect to {0}: {1}'.format(url, e))

    # Read timeouts
    except requests.Timeout as e:
        raise ApiError(0, 'Timed out reading from {0}: {1}'.format(url, e))

    # Try to get some response JSON data
    response_data = {}

    try:
        response_data = response.json()

    except ValueError:
        raise ApiError(
            response.status_code,
            'Invalid JSON response',
            content=response.content,
        )

    # Capture and re-raise any HTTP/API errors
    try:
        response.raise_for_status()

    except requests.HTTPError as e:
        raise ApiError(
            response.status_code,
            response_data.get('error_name', 'Unknown'),
            message=response_data.get('error_message', e),
            content=response_data,
        )

    return response_data


def ping(settings):
    return make_api_request(
        get_session().get,
        'server/{0}/ping'.format(settings.server_id),
        settings=settings,
    )


def _upload_states_return_settings(url, states, settings, json=None):
    json = json or states

    if json is not states:
        json['states'] = states

    response_data = make_api_request(
        get_session().post, url,
        settings=settings,
        json=json,
        # Explicitly set the max (matching server) timeout for syncing states
        # to avoid any sync "thrashing".
        timeout=600,
    )

    return response_data['settings']


def sync_states(states, settings):
    '''
    Uploads a full state to api.servicecanary.com and returns any data sent back
    from the server (settings).
    '''

    return _upload_states_return_settings(
        'server/{0}/sync'.format(settings.server_id),
        make_sync_or_changes_dict(states),
        settings,
        json={
            'hostname': socket.gethostname(),
            'canaryd_version': __version__,
        },
    )


def upload_state_changes(state_changes, settings):
    '''
    Uploads partial state to api.servicecanary.com.
    '''

    return _upload_states_return_settings(
        'server/{0}/state'.format(settings.server_id),
        make_sync_or_changes_dict(state_changes),
        settings,
    )


def create_event(settings, plugin, type, description, data=None):
    return make_api_request(
        get_session().post,
        'server/{0}/event'.format(settings.server_id),
        settings=settings,
        json={
            'plugin': plugin,
            'type': type,
            'description': description,
            'data': data,
        },
    )


def register(key):
    '''
    Register this server on api.servicecanary.com and return a server id.
    '''

    response_data = make_api_request(
        get_session().post,
        'servers',
        api_key=key,
        json={
            'hostname': socket.gethostname(),
            'canaryd_version': __version__,
        },
    )

    return response_data['server_id']


def signup(email):
    '''
    Signup for api.servicecanary.com and return the access key for that user.
    '''

    response_data = make_api_request(
        get_session().post,
        'signup',
        json={
            'email': email,
        },
    )

    if 'api_key' in response_data:
        return True, response_data['api_key']

    return False, response_data['info']



import socket

from distutils.spawn import find_executable
from os import path, sep as os_sep

from canaryd_packages import six

from canaryd.plugin import Plugin
from canaryd.subprocess import CalledProcessError

from .services_util import (
    get_initd_services,
    get_launchd_services,
    get_pid_to_listens,
    get_supervisor_services,
    get_systemd_services,
    get_upstart_services,
)

COMMAND_TO_FUNC = {
    'launchctl': get_launchd_services,
    'systemctl': get_systemd_services,
    'initctl': get_upstart_services,

    'supervisorctl': get_supervisor_services,
}


def check_port(ip_type, host, port):
    # Open our IPv4 or IPv6 socket
    socket_type = socket.AF_INET if ip_type == 'ipv4' else socket.AF_INET6
    sock = socket.socket(socket_type, socket.SOCK_STREAM)
    sock.settimeout(1)

    # If listening everywhere, just try localhost
    if host == '*':
        if ip_type == 'ipv4':
            host = '127.0.0.1'
        else:
            host = '::1'

    try:
        result = sock.connect_ex((host, port))
        return result == 0

    except (socket.error, socket.gaierror):
        pass

    finally:
        sock.close()


def make_service_data(data):
    blank_service = {
        'ports': [],
        'up_ports': [],
    }

    blank_service.update(data)
    return blank_service


class Services(Plugin):
    '''
    The services plugin provides a combined view of "system" services - ie the
    merge of any running launchctl/rcd/initd/upstart/systemd services.
    '''

    spec = ('service', {
        'running': bool,
        'pid': int,
        'enabled': bool,
        'init_system': six.text_type,
        'ports': [int],
        'up_ports': [int],
    })

    @staticmethod
    def prepare(settings):
        commands = COMMAND_TO_FUNC.keys()

        if not path.exists(path.join(os_sep, 'etc', 'init.d')) and not any(
            find_executable(command)
            for command in commands
        ):
            raise OSError('No container commands found: {0}'.format(commands))

    def get_state(self, settings):
        services = {}
        timeout = self.get_timeout(settings)

        for command, func in six.iteritems(COMMAND_TO_FUNC):
            if find_executable(command):
                services.update(func(timeout=timeout))

        if path.exists(path.join(os_sep, 'etc', 'init.d')):
            services.update(
                # Pass existing services to avoid overhead of running all the
                # init.d status scripts.
                get_initd_services(existing_services=services, timeout=timeout),
            )

        # Get mapping of PID -> listening ports
        try:
            pid_to_listens = get_pid_to_listens(timeout=timeout)

        except (CalledProcessError, OSError):
            pass

        else:
            # Augment services with their ports
            for name, data in six.iteritems(services):
                if 'pid' not in data or data['pid'] not in pid_to_listens:
                    data['ports'] = set()
                    data['up_ports'] = set()
                    continue

                data['ports'] = set(
                    port
                    for _, _, port in pid_to_listens[data['pid']]
                )

                data['up_ports'] = set(
                    port for ip_type, host, port in pid_to_listens[data['pid']]
                    if check_port(ip_type, host, port)
                )

        return dict(
            (key, make_service_data(data))
            for key, data in six.iteritems(services)
        )

    @staticmethod
    def get_change_key(change):
        # Ensure the running state is included in our change key, such that
        # if grouping we only group services that have started/stopped.
        return change.data and change.data.get('running')

    @staticmethod
    def should_apply_change(change):
        data_keys = list(six.iterkeys(change.data))

        # If only up_ports changes, we don't want to generate an update - we'll
        # generate any issues as needed below.
        if len(data_keys) == 1 and data_keys[0] == 'up_ports':
            return False

    @staticmethod
    def get_action_for_change(change):
        if change.type != 'updated':
            return

        if 'running' in change.data:
            was_running, _ = change.data['running']

            if was_running:
                return 'stopped'

            return 'started'

        if 'pid' in change.data:
            return 'restarted'

    @staticmethod
    def generate_issues_from_change(change, settings):
        key = change.key
        data_changes = change.data

        # If the script has been removed, resolve any leftover issues and exit
        # (the delete event is still created).
        if change.type == 'deleted':
            yield 'resolved', None, None
            return

        # If the service *was* running, but now isn't, resolve any leftover issues
        # as we assume ports are down intentionally.
        if 'running' in data_changes and data_changes['running'][0]:
            yield 'resolved', None, None
            return

        # Finally, check the ports!
        if 'up_ports' in data_changes:
            # If ports also changed, and changed to nothing, resolve anything
            # as the service must have been turned off.
            if 'ports' in data_changes and not data_changes['ports'][1]:
                yield 'resolved', None, None
                return

            from_ports, to_ports = data_changes['up_ports']

            # No ports up now?
            if from_ports and not to_ports:
                if settings.service_critical:
                    yield (
                        'critical',
                        'All {0} ports down'.format(key),
                        data_changes,
                    )

            # We lost 1+ port, but not all?
            elif from_ports and len(to_ports) < len(from_ports):
                if settings.service_warning:
                    yield (
                        'warning',
                        'Some {0} ports down'.format(key),
                        data_changes,
                    )

            # We have more ports than before? Assume resolved
            # TODO: improve this
            elif to_ports:
                yield 'resolved', 'All {0} ports up'.format(key), data_changes



from canaryd_packages import six

from canaryd.plugin import Plugin

# Mapping for iptables code arguments to variable names
IPTABLES_ARGS = {
    '-A': 'chain',
    '-j': 'jump',

    # Boolean matches
    '-p': 'protocol',
    '-s': 'source',
    '-d': 'destination',
    '-i': 'in_interface',
    '-o': 'out_interface',

    # Logging
    '--log-prefix': 'log_prefix',

    # NAT exit rules
    '--to-destination': 'to_destination',
    '--to-source': 'to_source',
    '--to-ports': 'to_ports',
}


class Iptables(Plugin):
    spec = ('chain', {
        'policy': six.text_type,
        'rules': [dict],
    })

    command = 'iptables-save'

    @staticmethod
    def parse(output):
        chains = {}

        for line in output.splitlines():
            # Parse the chains
            if line.startswith(':'):
                line = line[1:]

                chain, policy, _ = line.split()
                chains[chain] = {
                    'rules': [],
                    'policy': policy,
                }

            # Pass the rules
            if not line.startswith('-A'):
                continue

            bits = line.split()

            definition = {}

            key = None
            args = []
            not_arg = False

            def add_args():
                arg_string = ' '.join(args)

                if key in IPTABLES_ARGS:
                    definition_key = (
                        'not_{0}'.format(IPTABLES_ARGS[key])
                        if not_arg
                        else IPTABLES_ARGS[key]
                    )
                    definition[definition_key] = arg_string
                else:
                    definition.setdefault('extras', []).extend((key, arg_string))

            for bit in bits:
                if bit == '!':
                    if key:
                        add_args()
                        args = []
                        key = None

                    not_arg = True

                elif bit.startswith('-'):
                    if key:
                        add_args()
                        args = []
                        not_arg = False

                    key = bit

                else:
                    args.append(bit)

            if key:
                add_args()

            if 'extras' in definition:
                definition['extras'] = definition['extras']

            chain = definition.pop('chain')
            chains[chain]['rules'].append(definition)

        return chains

    @staticmethod
    def get_action_for_change(change):
        # If we change rules only, make the action "rules updated|addded|deleted"
        if 'rules' in change.data and len(change.data) == 1:
            return 'rules {0}'.format(change.type)


class Ip6tables(Iptables):
    command = 'ip6tables-save'



# canaryd
# File: canaryd/__main__.py
# Desc: entry point for canaryd

import logging

from time import time

from canaryd_packages import click

from canaryd.daemon import run_daemon
from canaryd.log import logger, setup_logging, setup_logging_from_settings
from canaryd.plugin import (
    get_and_prepare_working_plugins,
    get_plugin_states,
)
from canaryd.remote import backoff, ping, sync_states
from canaryd.settings import ensure_config_directory, get_settings
from canaryd.version import __version__


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.option('-v', '--verbose', is_flag=True)
@click.option('-d', '--debug', is_flag=True)
@click.version_option(
    version=__version__,
    prog_name='canaryd',
    message='%(prog)s: v%(version)s',
)
def main(verbose, debug):
    '''
    Run the canaryd daemon.
    '''

    log_level = setup_logging(verbose, debug)

    logger.info('Starting canaryd v{0}'.format(__version__))
    logger.info('Log level set to: {0}'.format(
        logging.getLevelName(log_level),
    ))

    # Ensure the config directory exists
    ensure_config_directory()

    # Load the settings, using our config file if provided
    settings = get_settings()

    # Setup any log file/syslog
    setup_logging_from_settings(settings)

    # Settings can set debug on if needed
    if settings.debug == 'true':
        logger.setLevel(logging.DEBUG)

    if not settings.api_key or not settings.server_id:
        logger.critical('Missing api_key and/or server_id in config file')
        return

    # Initial ping for API presence
    logger.info('Ping API...')
    backoff(
        ping, settings,
        error_message='Could not ping',
        max_wait=settings.collect_interval_s,
    )

    # Load the plugin list
    plugins = get_and_prepare_working_plugins(settings)

    # Get the initial state
    logger.info('Getting initial state...')
    start_time = time()
    states = get_plugin_states(plugins, settings)

    # Sync this state and get settings
    logger.info('Syncing initial state...')

    remote_settings = backoff(
        sync_states, states, settings,
        error_message='Could not sync state',
        max_wait=settings.collect_interval_s,
    )

    # Update settings w/remote ones
    settings.update(remote_settings)

    # Run the loop
    logger.info('Starting daemon loop...')

    # Make previous states dict
    previous_states = dict(
        (plugin, status_data[1])
        for plugin, status_data in states
    )

    run_daemon(previous_states, settings, start_time=start_time)


try:
    main()

except Exception:
    # TODO: public Sentry logging

    raise



import re

from canaryd_packages import six

from canaryd.subprocess import get_command_output

DEB_REGEX = r'^[a-z]+\s+([a-zA-Z0-9\+\-\.]+):?[a-zA-Z0-9]*\s+([a-zA-Z0-9:~\.\-\+]+).+$'
RPM_REGEX = r'^([a-zA-Z0-9_\-\+]+)\-([0-9a-z\.\-]+)\.[a-z0-9_\.]+$'
PKG_REGEX = r'^([a-zA-Z0-9_\-\+]+)\-([0-9a-z\.]+)'


def get_parse_packages(timeout, package_type, command, regex, lower=True):
    output = get_command_output(
        command,
        timeout=timeout,
    )

    packages = {}

    for line in output.splitlines():
        matches = re.match(regex, line)

        if matches:
            # Sort out name
            name = matches.group(1)
            if lower:
                name = name.lower()

            packages.setdefault(name, set())
            packages[name].add(matches.group(2))

    return dict(
        (key, {
            'versions': list(value),
            'package_type': package_type,
        })
        for key, value in six.iteritems(packages)
    )


def get_deb_packages(timeout):
    return get_parse_packages(timeout, 'deb', 'dpkg -l', DEB_REGEX)


def get_rpm_packages(timeout):
    return get_parse_packages(timeout, 'rpm', 'rpm -qa', RPM_REGEX)


def get_pkg_packages(timeout):
    return get_parse_packages(timeout, 'pkg', 'pkg_info', PKG_REGEX)



# canaryd
# File: canaryd/daemon.py
# Desc: the canaryd daemon

from time import sleep, time

from canaryd.diff import get_state_diff
from canaryd.log import logger
from canaryd.plugin import (
    get_and_prepare_working_plugins,
    get_plugin_states,
)
from canaryd.remote import backoff, upload_state_changes


def _sleep_until_interval(start, interval):
    time_taken = time() - start

    if time_taken < interval:
        sleep(interval - time_taken)


def _daemon_loop(previous_states, settings):
    logger.info('Getting plugin states...')

    # Load the plugin list
    plugins = get_and_prepare_working_plugins(settings)

    states = get_plugin_states(plugins, settings)
    state_changes = []

    for plugin, status_data in states:
        status, data = status_data

        # Plugin ran OK and we have state!
        if status:
            previous_state = previous_states.get(plugin, {})

            # If the previous state was good - ie not an Exception instance - this
            # prevents first-failing then working plugin from generating addition
            # events on first successful run.
            if isinstance(previous_state, dict):
                state_diff = get_state_diff(plugin, data, previous_state)
                state_changes.append((plugin, (status, state_diff)))

            # Because we don't know the previous working state, send the whole
            # state obj to the server to diff, like the initial sync.
            else:
                state_changes.append((plugin, ('SYNC', data)))

            # Set the previous state
            previous_states[plugin] = data

        # Plugin raised an exception, fail!
        else:
            logger.critical((
                'Unexpected exception while getting {0} state: '
                '{1}({2})'
            ).format(plugin.name, data.__class__.__name__, data))

            # Send the failed exception to the server, generating a warning
            exception_data = {
                'class': data.__class__.__name__,
                'message': '{0}'.format(data),
                'traceback': getattr(data, '_traceback'),
            }
            state_changes.append((plugin, ('ERROR', exception_data)))

    logger.info('Uploading state changes...')

    settings_changes = backoff(
        upload_state_changes,
        state_changes,
        settings,
        error_message='Could not sync state changes',
        max_wait=settings.collect_interval_s,
    )

    if settings_changes:
        changed_keys = settings.update(settings_changes)
        if changed_keys:
            logger.info('Remote settings update: {0}'.format(
                ', '.join(changed_keys),
            ))


def run_daemon(previous_states, settings, start_time=None):
    if start_time:
        _sleep_until_interval(
            start_time, settings.collect_interval_s,
        )

    while True:
        start = time()

        _daemon_loop(previous_states, settings)

        _sleep_until_interval(
            start, settings.collect_interval_s,
        )



from distutils.spawn import find_executable

from canaryd_packages import six

from canaryd.plugin import Plugin

from .packages_util import (
    get_deb_packages,
    get_pkg_packages,
    get_rpm_packages,
)

COMMAND_TO_FUNC = {
    'dpkg': get_deb_packages,
    'rpm': get_rpm_packages,
    'pkg_info': get_pkg_packages,
}


class Packages(Plugin):
    spec = ('package', {
        'versions': [six.text_type],
        'package_type': six.text_type,
    })

    @staticmethod
    def prepare(settings):
        commands = COMMAND_TO_FUNC.keys()

        if not any(
            find_executable(command)
            for command in commands
        ):
            raise OSError('No container commands found: {0}'.format(commands))

    def get_state(self, settings):
        packages = {}
        timeout = self.get_timeout(settings)

        for command, func in six.iteritems(COMMAND_TO_FUNC):
            if find_executable(command):
                packages.update(func(timeout=timeout))

        return packages



from canaryd_packages import six

from canaryd.plugin import Plugin

# We ignore these because they constantly update and represent "live" state
# rather than configuration, which is what we're trying to track here.
IGNORE_PREFIXES = ('vm.', 'machdep.', 'debug.')


class Sysctl(Plugin):
    spec = ('key', {
        'values': [six.text_type],
    })

    command = (
        '(cat /etc/sysctl.conf || true); '
        '(cat /etc/sysctl.d/*.conf || true); '
        '(cat /run/sysctl.d/*.conf || true); '
        '(cat /usr/lib/sysctl.d/*.conf || true)'
    )

    def parse(self, output):
        state = {}

        for name, values in self.parse_lines(output.splitlines()):
            state[name] = {
                'values': values,
            }

        return state

    @staticmethod
    def parse_lines(lines):
        for line in lines:
            line = line.strip()

            if (
                line.startswith('#') or
                any(line.startswith(prefix) for prefix in IGNORE_PREFIXES)
            ):
                continue

            if ':' in line:
                bits = line.split()
                if len(bits) <= 1:
                    continue

                bits = [bits[0][:-1], bits[1]]

            else:
                bits = line.split('=')

            if len(bits) <= 1:
                continue

            name, value = bits
            name = name.strip()
            values = [v.strip() for v in value.split()]

            yield name, values



from os import geteuid

from canaryd_packages import click

from canaryd.exceptions import CanarydError, UserCancelError


def check_root(message, exit=False):
    # If we're root, we're all good!
    if geteuid() <= 0:
        return

    message = click.style(message, 'yellow')

    # If exit, just fail
    if exit:
        raise CanarydError(message)

    # Warn the user about the implications of running as non-root
    click.echo(message)

    # Confirm this is OK!
    if not click.confirm('Do you wish to continue?'):
        raise UserCancelError()



# canaryd
# File: canaryd/settings.py
# Desc: settings for canaryd/canaryctl

import platform

from os import environ, geteuid, listdir, makedirs, path
from shutil import copy

from canaryd_packages import click, six
from canaryd_packages.six.moves.configparser import (
    DuplicateSectionError,
    Error as ConfigParserError,
    RawConfigParser,
)

from canaryd.exceptions import ConfigError
from canaryd.log import logger

API_BASE = environ.get(
    'API_BASE',
    'https://api.servicecanary.com',
)

API_VERSION = int(environ.get(
    'API_VERSION',
    1,
))

API_KEY = environ.get(
    'API_KEY',
    None,
)

SERVER_ID = environ.get(
    'SERVER_ID',
    None,
)


class CanarydSettings(object):
    api_base = API_BASE
    api_version = API_VERSION

    log_file = None
    # Rotate logs by this size (in bytes!) or a TimedRotatingFileHandler "when"
    log_file_rotation = None
    # Number of files to rotate
    log_file_rotation_count = 5

    # syslog facility to log to
    syslog_facility = None

    debug = False

    collect_interval_s = 30

    # API access key
    api_key = API_KEY

    # ID of (hopefully this) server attached to this key
    server_id = SERVER_ID

    def __init__(self, **kwargs):
        self.update(kwargs)

        # If no log file specified, we're root and /var/log exists, use that
        if (
            self.log_file is None
            and geteuid() <= 0
            and path.exists(path.join('/', 'var', 'log'))
        ):
            logger.debug('Root user, so setting log file to /var/log/canaryd.log')
            self.log_file = path.join('/', 'var', 'log', 'canaryd.log')

    def __getattr__(self, key):
        try:
            return super(CanarydSettings, self).__getattr__(key)
        except AttributeError:
            pass

    def update(self, data):
        changed_keys = []

        for key, value in six.iteritems(data):
            if getattr(self, key, None) != value:
                setattr(self, key, value)
                changed_keys.append(key)

        logger.debug('Settings updated: {0} <= {1}'.format(changed_keys, data))
        return changed_keys


def get_config_directory():
    # If we're non-root, just use the users config dir (~/.config, etc)
    if geteuid() > 0:
        config_directory = click.get_app_dir('canaryd')

        return config_directory

    # If we're root on OSX
    if platform.system() == 'Darwin':
        return path.join('/', 'Library', 'Application Support', 'canaryd')

    # Elsewhere on *nix (no Windows support _yet_)
    return path.join('/', 'etc', 'canaryd')


def get_config_file():
    return path.join(get_config_directory(), 'canaryd.conf')


def get_scripts_directory():
    return path.join(get_config_directory(), 'scripts')


def _get_settings(config_file=None):
    '''
    Load the config from the filesystem if provided, with defaults.
    '''

    config_file = config_file or get_config_file()

    settings = CanarydSettings()

    parser = RawConfigParser()

    if path.exists(config_file):
        try:
            parser.read(config_file)
            canaryd_config_items = parser.items('canaryd')

            settings.update(dict(canaryd_config_items))
            settings.update(dict(canaryd_config_items))

        except ConfigParserError as e:
            logger.critical('Error in config file ({0}): {1}'.format(
                config_file, e.message,
            ))
            raise ConfigError('Config file error')

    return settings


SETTINGS = None


def get_settings(config_file=None):
    '''
    Cached/public version of _get_settings.
    '''

    global SETTINGS

    if SETTINGS is None:
        SETTINGS = _get_settings(config_file=config_file)

    return SETTINGS


def write_settings_to_config(settings):
    '''
    Write a config file from settings.
    '''

    # Ensure the config directory exists
    ensure_config_directory()

    # Generate the config
    config = RawConfigParser()
    config_file = get_config_file()

    # Attempt to work alongside any existing config
    try:
        config.read(config_file)

    except ConfigParserError as e:
        logger.critical('Error in config file ({0}): {1}'.format(
            config_file, e.message,
        ))
        raise ConfigError('Config file error')

    try:
        config.add_section('canaryd')
    except DuplicateSectionError:
        pass

    for key, value in six.iteritems(settings.__dict__):
        config.set('canaryd', key, value)

    with open(config_file, 'wb') as f:
        config.write(f)


def copy_builtin_scripts():
    # Copy built in scripts to the scripts/available directory
    available_scripts_directory = path.join(get_scripts_directory(), 'available')

    logger.debug(
        'Copying default scripts to: {0}'.format(available_scripts_directory),
    )

    builtin_scripts_directory = path.join(path.dirname(__file__), 'scripts')

    for file in listdir(builtin_scripts_directory):
        copy(
            path.join(builtin_scripts_directory, file),
            path.join(available_scripts_directory, file),
        )


def ensure_config_directory():
    # Make sure the config directory exists
    config_directory = get_config_directory()

    if not path.exists(config_directory):
        logger.debug('Creating config directory: {0}'.format(config_directory))
        makedirs(config_directory)

    # Make sure the scripts directory exists
    scripts_directory = get_scripts_directory()

    if not path.exists(scripts_directory):
        logger.debug('Creating scripts directory: {0}'.format(scripts_directory))

        # Make the scripts, scripts/enabled & scripts/available directories
        makedirs(path.join(scripts_directory, 'enabled'))
        makedirs(path.join(scripts_directory, 'available'))

        copy_builtin_scripts()



from distutils.spawn import find_executable
from os import path
from pkgutil import get_data

from canaryd_packages import click


def write_script(script_path, script):
    # get_data returns binary, so make it a string
    script = script.decode('utf-8')

    # Swap in the location of canaryd
    canaryd_location = find_executable('canaryd')
    script = script.replace('CANARYD_LOCATION', canaryd_location)

    # Write the init system script
    with open(script_path, 'w') as file:
        file.write(script)


def install_service():
    script = None
    script_path = None
    start_command = None
    enable_command = None

    # OSX/Darwin
    if find_executable('launchctl'):
        script = get_data('canaryd', path.join('init_scripts', 'com.oxygem.canaryd.plist'))
        script_path = path.join('/', 'Library', 'LaunchDaemons', 'com.oxygem.canaryd.plist')
        start_command = 'launchctl load {0}'.format(script_path)
        enable_command = 'launchctl enable system/com.oxygem.canaryd'

    # Systemd
    elif find_executable('systemctl'):
        script = get_data('canaryd', path.join('init_scripts', 'canaryd.service'))
        script_path = path.join('/', 'etc', 'systemd', 'system', 'canaryd.service')
        start_command = 'systemctl start canaryd.service'
        enable_command = 'systemctl enable canaryd.service'

    # Upstart
    elif find_executable('initctl'):
        script = get_data('canaryd', path.join('init_scripts', 'canaryd.conf'))
        script_path = path.join('/', 'etc', 'init', 'canaryd.conf')
        start_command = 'initctl start canaryd'

    # Init.d
    elif path.exists(path.join('/', 'etc', 'init.d')):
        script = get_data('canaryd', path.join('init_scripts', 'canaryd.sh'))
        script_path = path.join('/', 'etc', 'init.d', 'canaryd')
        start_command = 'chmod +x /etc/init.d/canaryd && /etc/init.d/canaryd start'

        if find_executable('update-rc.d'):
            enable_command = 'update-rc.d canaryd defaults'

        elif find_executable('chkconfig'):
            enable_command = 'chkconfig canaryd on'

    write_script(script_path, script)
    click.echo('--> {0} written'.format(script_path))

    return start_command, enable_command



from os import listdir, path, remove, symlink

from canaryd.settings import get_scripts_directory

INVALID_EXTENSIONS = (
    '.pyc',
)


class ScriptException(Exception):
    pass


class NoScriptFoundError(ScriptException):
    '''
    Raised when a script file cannot be located.
    '''


class NoScriptChangesError(ScriptException):
    '''
    Raised when no changes are made when enabling/disabling scripts.
    '''


class ScriptNotLinkError(ScriptException):
    '''
    Raised when the script is not a link when disabling scripts.
    '''


def get_scripts(settings):
    available_scripts = listdir(path.join(get_scripts_directory(), 'available'))
    enabled_scripts = listdir(path.join(get_scripts_directory(), 'enabled'))

    all_scripts = set(available_scripts + enabled_scripts)
    all_scripts = filter(
        lambda script: not script.endswith(INVALID_EXTENSIONS),
        all_scripts,
    )

    return (
        (script, script in enabled_scripts)
        for script in all_scripts
    )


def enable_script(script):
    scripts_directory = get_scripts_directory()

    source_script = path.join(scripts_directory, 'available', script)
    link_name = path.join(scripts_directory, 'enabled', script)

    if not path.exists(source_script):
        raise NoScriptFoundError(source_script)

    if path.exists(link_name):
        raise NoScriptChangesError()

    symlink(
        source_script,
        link_name,
    )


def disable_script(script):
    scripts_directory = get_scripts_directory()

    source_script = path.join(scripts_directory, 'available', script)
    link_name = path.join(scripts_directory, 'enabled', script)

    if not path.exists(source_script):
        raise NoScriptFoundError(source_script)

    if not path.exists(link_name):
        raise NoScriptChangesError()

    if not path.islink(link_name):
        raise ScriptNotLinkError(
            'Script {0} is not a link. You should move it to: {1}.'.format(
                link_name, source_script,
            ),
        )

    remove(link_name)



from canaryd_packages import click

from canaryd.exceptions import CanarydError
from canaryd.script import (
    disable_script,
    enable_script,
    get_scripts,
    NoScriptChangesError,
    NoScriptFoundError,
    ScriptNotLinkError,
)
from canaryd.settings import (
    copy_builtin_scripts,
    get_settings,
)

from . import main


@main.group(invoke_without_command=True)
@click.pass_context
def scripts(ctx):
    '''
    List and manage scripts for canaryd.

    \b
    # List scripts
    canaryctl scripts

    \b
    # Enable a script
    canaryctl scripts enable <script.sh>

    \b
    # Disable a script
    canaryctl scripts disable <script.sh>
    '''

    if ctx.invoked_subcommand is not None:
        return

    scripts = get_scripts(get_settings())

    click.echo('--> Scripts:')

    for script in scripts:
        click.echo('    {0}, enabled: {1}'.format(
            click.style(script[0], bold=True),
            script[1],
        ))


@scripts.command()
def copy():
    '''
    Copy the builtin scripts into the servers canaryd settings directory.
    '''

    copy_builtin_scripts()


@scripts.command()
@click.argument('script')
def enable(script):
    '''
    Enable a script.
    '''

    try:
        enable_script(script)

    except NoScriptFoundError as e:
        raise CanarydError(click.style(
            'No script file ({0}) exists.'.format(e.message),
            'red',
        ))

    except NoScriptChangesError:
        click.echo('Script {0} is already enabled.'.format(
            click.style(script, bold=True),
        ))
        return

    click.echo('Script enabled: {0}'.format(click.style(script, bold=True)))


@scripts.command()
@click.argument('script')
def disable(script):
    '''
    Disable a script.
    '''

    try:
        disable_script(script)

    except NoScriptFoundError as e:
        raise CanarydError(click.style(
            'No script file ({0}) exists.'.format(e.message),
            'red',
        ))

    except ScriptNotLinkError as e:
        raise CanarydError(click.style(e.message, 'yellow'))

    except NoScriptChangesError:
        click.echo('Script {0} is already disabled.'.format(
            click.style(script, bold=True),
        ))
        return

    click.echo('Script disabled: {0}'.format(click.style(script, bold=True)))



from distutils.spawn import find_executable

from canaryd_packages import six

from canaryd.plugin import Plugin

from .containers_util import (
    get_docker_containers,
    get_lxc_containers,
    get_openvz_containers,
    get_virsh_containers,
)

COMMAND_TO_FUNC = {
    'docker': get_docker_containers,
    'lxc': get_lxc_containers,
    'vzlist': get_openvz_containers,
    'virsh': get_virsh_containers,
}


def make_container_data(data):
    blank_container = {
        'environment': [],
        'names': [],
        'ips': [],
    }

    blank_container.update(data)
    return blank_container


class Containers(Plugin):
    spec = ('container', {
        'runtime': six.text_type,
        'running': bool,
        'pid': int,
        'command': six.text_type,
        'image': six.text_type,
        'id': six.text_type,
        'environment': [six.text_type],
        'names': [six.text_type],
        'ips': [six.text_type],
    })

    @staticmethod
    def prepare(settings):
        commands = COMMAND_TO_FUNC.keys()

        if not any(
            find_executable(command)
            for command in commands
        ):
            raise OSError('No container commands found: {0}'.format(commands))

    def get_state(self, settings):
        containers = {}
        timeout = self.get_timeout(settings)

        for command, func in six.iteritems(COMMAND_TO_FUNC):
            if find_executable(command):
                containers.update(func(timeout=timeout))

        return dict(
            (key, make_container_data(data))
            for key, data in six.iteritems(containers)
        )

    @staticmethod
    def get_action_for_change(change):
        if change.type != 'updated':
            return

        if 'running' in change.data:
            was_running, _ = change.data['running']

            if was_running:
                return 'stopped'

            return 'started'

        if 'pid' in change.data:
            return 'restarted'



import json

from canaryd.subprocess import CalledProcessError, get_command_output


def get_docker_containers(timeout):
    containers = {}

    try:
        output = get_command_output(
            'docker inspect `docker ps -qa`',
            shell=True,
            timeout=timeout,
        )

    # Either Docker is down or there are no containers
    except CalledProcessError:
        return containers

    data = json.loads(output)

    for container in data:
        container_data = {
            'runtime': 'docker',
            'running': container['State']['Running'],
            'command': ' '.join(container['Config']['Cmd']),
            'environment': container['Config']['Env'],
            'image': container['Config']['Image'],
            'id': container['Id'],
        }

        # Figure out the name(s)
        if 'Names' in container:
            container_data['names'] = container['Names']

        elif 'Name' in container:
            container_data['names'] = [container['Name']]

        # If running, get the PID
        if container_data['running']:
            container_data['pid'] = container['State']['Pid']

        container_id = container['Id'][:6]
        container_key = 'docker/{0}'.format(container_id)

        containers[container_key] = container_data

    return containers


def get_lxc_containers(timeout):
    output = get_command_output(
        'lxc list --fast --format json',
        timeout=timeout,
    )

    data = json.loads(output)
    containers = {}

    for container in data:
        container_data = {
            'runtime': 'lxc',
            'names': [container['name']],
            'running': container['status'] == 'Running',
        }

        container_key = 'lxc/{0}'.format(container['name'])
        containers[container_key] = container_data

    return containers


def get_openvz_containers(timeout):
    output = get_command_output(
        'vzlist -a -j',
        timeout=timeout,
    )
    combined_json = ''.join(output)

    data = json.loads(combined_json)
    containers = {}

    for container in data:
        container_data = {
            'runtime': 'openvz',
            'ips': container['ip'],
            'image': container['ostemplate'],
            'running': container['status'] == 'running',
        }

        container_key = 'openvz/{0}'.format(container['ctid'])
        containers[container_key] = container_data

    return containers


def get_virsh_containers(timeout):
    output = get_command_output(
        'virsh list --all',
        timeout=timeout,
    )
    containers = {}

    for line in output.splitlines()[2:]:
        bits = line.split()

        if len(bits) != 3:
            continue

        id, name, state = bits

        container_data = {
            'runtime': 'virsh',
            'names': [name],
            'running': state == 'running',
        }

        container_key = 'virsh/{0}'.format(name)
        containers[container_key] = container_data

    return containers



import re

from collections import defaultdict
from os import listdir, path, sep as os_sep

from canaryd.subprocess import CalledProcessError, get_command_output

# We ignore these as they regularly get deleted/added as part of normal OSX
# lifecycle - and as such any events generated are not of use.
LAUNCHCTL_IGNORE_NAMES = ('oneshot', 'mdworker', 'mbfloagent')

# Systemd service types to ignore
SYSTEMD_IGNORE_TYPES = ('oneshot',)
SYSTEMD_REGEX = r'^([a-z\-]+)\.service\s+[a-z\-]+\s+[a-z]+\s+([a-z]+)'

UPSTART_REGEX = r'^([a-z\-]+) [a-z]+\/([a-z]+),?\s?(process)?\s?([0-9]+)?'
SUPERVISOR_REGEX = r'([a-z\-]+)\s+([A-Z]+)\s+pid\s([0-9]+)'

INITD_REGEX = r'([a-zA-Z0-9\-]+)=([0-9]+)=([0-9]+)?'
INITD_USAGE_REGEX = re.compile(r'Usage:[^\n]+status')
IGNORE_INIT_SCRIPTS = []

# We *require* procfs to check PID -> port mappings
HAS_PROCFS = path.exists('/proc')
PID_TO_PORTS = defaultdict(set)


def get_pid_to_listens(timeout):
    if not HAS_PROCFS:
        return PID_TO_PORTS

    # Loop through the results and cleanup any PIDs that don't exist
    for pid in list(PID_TO_PORTS.keys()):
        if not path.exists('/proc/{0}'.format(pid)):
            PID_TO_PORTS.pop(pid)

    # Now get/update our PID -> port mapping
    output = get_command_output(
        'lsof -i -n -P -s TCP:LISTEN',
        timeout=timeout,
    )

    for line in output.splitlines():
        bits = line.split(None, 9)

        if bits[-1] != '(LISTEN)':
            continue

        _, pid, _, _, ip_type, _, _, _, ip_host, _ = bits

        ip_type = ip_type.lower()
        pid = int(pid)

        # Work out the host:port bit
        host, port = ip_host.rsplit(':', 1)
        port = int(port)

        host_port_tuple = (ip_type, host, port)
        PID_TO_PORTS[pid].add(host_port_tuple)

    return PID_TO_PORTS


def get_launchd_services(timeout):
    '''
    Execute & parse ``launchctl list``.
    '''

    output = get_command_output(
        'launchctl list',
        timeout=timeout,
    )

    services = {}

    for line in output.splitlines():
        if any(name in line for name in LAUNCHCTL_IGNORE_NAMES):
            continue

        bits = line.split()

        if not bits or bits[0] == 'PID':
            continue

        name = bits[2]

        # If the last "bit" is just a number, it's the PID, so we strip it
        name_bits = name.split('.')
        if name_bits[-1].isdigit():
            name = '.'.join(name_bits[:-1])

        data = {
            'init_system': 'launchd',
        }

        try:
            data['pid'] = int(bits[0])
        except ValueError:
            pass

        data['running'] = 'pid' in data

        services[name] = data

    return services


def get_initd_services(existing_services, timeout):
    existing_services = existing_services or []

    init_dir = path.join(os_sep, 'etc', 'init.d')
    service_names = listdir(init_dir)
    services = {}

    for name in service_names:
        if name in existing_services or name in IGNORE_INIT_SCRIPTS:
            continue

        script_path = path.join(init_dir, name)

        with open(script_path) as script:
            script_data = script.read()

            if (
                # LSB header
                '### BEGIN INIT INFO' not in script_data
                or not INITD_USAGE_REGEX.search(script_data)
            ):
                IGNORE_INIT_SCRIPTS.append(name)
                continue

        # Get the status
        status = False

        try:
            get_command_output(
                (script_path, 'status'),
                timeout=timeout,
            )
            status = True

        except (OSError, CalledProcessError):
            pass

        # Get the PID
        pid = None
        pid_line = None

        try:
            pid_line = get_command_output(
                '''
                ps --ppid 1 -o 'tty,pid,comm' | \
                grep '^\?.*\s{0}.*$' | \
                head -n 1
                '''.format(name),
                shell=True,
                timeout=timeout,
            )
        except CalledProcessError:
            raise

        if pid_line:
            bits = pid_line.strip().split()
            try:
                pid = int(bits[-2])
            except (TypeError, ValueError):
                raise

        # Check if enabled
        enabled = False

        try:
            found_links = get_command_output(
                'find /etc/rc*.d/S*{0} -type l'.format(name),
                timeout=timeout,
            )

            if found_links.strip():
                enabled = True

        except CalledProcessError:
            pass

        services[name] = {
            'running': status or isinstance(pid, int),
            'pid': pid,
            'enabled': enabled,
            'init_system': 'initd',
        }

    return services


def get_systemd_services(timeout):
    output = get_command_output(
        'systemctl -alt service list-units',
        timeout=timeout,
    )

    services = {}

    for line in output.splitlines():
        line = line.strip()
        matches = re.match(SYSTEMD_REGEX, line)
        if matches:
            name = matches.group(1)

            # Get service info to extract pid/enabled status
            service_output = get_command_output(
                'systemctl show {0}.service'.format(name),
            )

            service_meta = {}
            for line in service_output.splitlines():
                key, value = line.split('=', 1)
                service_meta[key] = value

            if service_meta.get('Type') in SYSTEMD_IGNORE_TYPES:
                continue

            pid = service_meta.get(
                'ExecMainPID',
                service_meta.get('MainPID', None),
            )

            if pid:
                pid = int(pid)

            enabled = service_meta.get('UnitFileState') == 'enabled'

            services[name] = {
                'running': matches.group(2) == 'running',
                'pid': pid,
                'enabled': enabled,
                'init_system': 'systemd',
            }

    return services


def get_upstart_services(timeout):
    output = get_command_output(
        'initctl list',
        timeout=timeout,
    )

    services = {}

    for line in output.splitlines():
        matches = re.match(UPSTART_REGEX, line)
        if matches:
            name = matches.group(1)
            pid = matches.group(4)

            if pid:
                pid = int(pid)

            enabled = True

            # Check if enabled by looking in any override file
            override_filename = path.join(
                os_sep, 'etc', 'init',
                '{0}.override'.format(name),
            )
            if path.exists(override_filename):
                with open(override_filename) as script:
                    if 'manual' in script.read():
                        enabled = False

            services[name] = {
                'running': matches.group(2) == 'running',
                'pid': pid,
                'enabled': enabled,
                'init_system': 'upstart',
            }

    return services


def get_supervisor_services(timeout):
    output = get_command_output(
        'supervisorctl status',
        timeout=timeout,
    )

    services = {}

    for line in output.splitlines():
        matches = re.match(SUPERVISOR_REGEX, line)

        if matches:
            name = matches.group(1)
            status = matches.group(2)
            pid = matches.group(3)

            services[name] = {
                'running': status == 'RUNNING',
                'pid': int(pid),
                'init_system': 'supervisor',
            }

    return services
